"""
    MDAViscosity

Modal Decay Averaged Artificial Viscosity model, cf. Yu, Hesthaven, 2019.
"""
Base.@kwdef struct MDAViscosity <: AbstractArtificialViscosity
  mesh::Mesh
  cmax::Float64 = 1.0
  # buffers
  uhat::Matrix{Float64}
  vhat::Vector{Float64}
  lsq_x::Vector{Float64}
  lsq_y::Vector{Float64}
  mod_uh::Vector{Float64}
  skymod_uh::Vector{Float64}
  cellnorm2_u::Vector{Float64}
end


function MDAViscosity(mesh, cmax)
  Np, K = layout(mesh)
  N = Np-1
  # setup buffers
  uhat = zeros(Float64, layout(mesh))
  vhat = collect(Float64, 1:N).^(-N)
  vhat ./= norm(vhat)
  lsq_x = @. - log(1:N)
  lsq_y = zeros(Float64, N)
  mod_uh = zeros(Float64, N)
  skymod_uh = zeros(Float64, N)
  cellnorm2_u = zeros(Float64, K)
  return MDAViscosity(mesh, cmax, uhat, vhat, lsq_x, lsq_y, mod_uh, skymod_uh, cellnorm2_u)
end


function compute_viscosity!(
    mu,
    u, v,
    av::MDAViscosity)

  @unpack mesh, cmax, uhat, vhat, lsq_x, lsq_y,
          mod_uh, skymod_uh, cellnorm2_u = av
  @unpack element, K, L, x = mesh
  @unpack invV, N = element

  # h = L / (K * N)
  h = L / K

  u = reshape(u, layout(mesh))
  mul!(uhat, invV, u)

  dg1d.cellwise_inner_product!(cellnorm2_u, u, u, mesh)
  map!(cellnorm2_u, cellnorm2_u) do c
    abs(c) < 1e-10 ? 1.0 : c
  end

  maxabs_v = mapreduce(abs, max, v)
  fill!(mu, cmax * h / N * maxabs_v)

  # estimate effective decay rate by least square fit of modal coefficients, (3.9)
  # sum_m ( log(|q_m|) - (log(c) - tau log(m)) )^2 = sum_m (y_m - (a + b x_m))^2 -> min
  # log(m) ... grid points in log scale

  for j in 1:K
    # (3.11)
    # only consider coefficients 2:Np (which corresponds to 1:N), because
    # the 1st (0th) one corresponds to a constant mode that is irrelevant
    # for the smoothness esimation.
    @. mod_uh = sqrt( @views(uhat[2:end,j])^2 + cellnorm2_u[j] * vhat^2 )
    # (3.12)
    @views for n = 1:N
      skymod_uh[n] = maximum(mod_uh[min(n,N-1):end])
    end

    # data points in log scale
    @. lsq_y = log(skymod_uh)
    map!(lsq_y, lsq_y) do y
      isinf(y) ? - 20.0 : y
    end
    _, _, tau = dg1d.least_sqr_linear(lsq_x, lsq_y)
    isnan(tau) && error("undesired nan encountered")

    if 1.0 <= tau <= 3.0
      mu[j] *= 1.0 - 0.5 * (tau - 1.0)
    elseif tau > 3.0
      mu[j] *= 0.0
    end
  end

  return true
end
