function make_Equation(eos, prms)
  @unpack atm_density, atm_threshold_factor = prms
  atm = Atmosphere(atm_density, atm_threshold_factor)
  cons2prim_newtonsolver = dg1d.NewtonSolver(approx_derivative=true)
  cons2prim_bisection = dg1d.Bisection()
  return Equation(eos, atm, cons2prim_newtonsolver, cons2prim_bisection)
end


#######################################################################
#                           Implementation                            #
#######################################################################


module Formulae
  # auxilary variables: W, h, s
  Wprim(ρ, v, ϵ, p) = 1.0 / sqrt(1.0 - v^2) # in terms of primitive variables
  Wcons(D, S, τ, p) = (τ + p + D) / sqrt((τ + p + D)^2 - S^2) # in terms of conservative variables
  # inverse relations, see Bugner et al. (2018), see my notes
  ρ(D, S, τ, p) = D / Wcons(D, S, τ, p)
  v(D, S, τ, p) = S / (τ + D + p)
  ϵ(D, S, τ, p) = (sqrt((τ + p + D)^2 - S^2) - Wcons(D, S, τ, p) * p) / D - 1
  # conservative variables D, S, τ in terms of primitive ones
  D(ρ, v, ϵ, p) = ρ * Wprim(ρ, v, ϵ, p)
  S(ρ, v, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(ρ, v, ϵ, p)^2 * v
  τ(ρ, v, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(ρ, v, ϵ, p)^2 - p - D(ρ, v, ϵ, p)
  # conservative fluxes
  flux_D(ρ, v, ϵ, p) = D(ρ, v, ϵ, p) * v
  flux_S(ρ, v, ϵ, p) = S(ρ, v, ϵ, p) * v + p
  flux_τ(ρ, v, ϵ, p) = (τ(ρ, v, ϵ, p) + p) * v
end


@with_signature function flux(equation::Equation)
  @accepts State(D, S, tau), p, rho, v, eps
  flx_D   = Formulae.flux_D(rho, v, eps, p)
  flx_S   = Formulae.flux_S(rho, v, eps, p)
  flx_tau = Formulae.flux_τ(rho, v, eps, p)
  @returns flx_D, flx_S, flx_tau
end


@with_signature function speed(equation::Equation)
  @accepts State(D, S, tau), p, rho, v, eps
  @unpack eos = equation
  cs2 = eos(SquareSpeedOfSound, Density, rho, InternalEnergy, eps)
  cs = sqrt(abs(cs2)) # abs because cs2 might be negative due to cons2prim_insane
  v0 = v
  vp = (v + cs) / (1 + v * cs)
  vm = (v - cs) / (1 - v * cs)
  max_v = maximum(abs.((v0, vp, vm)))
  @returns max_v
end


# @with_signature function impose_atmosphere(equation::Equation)
#   @accepts rho, p, eps, v, D, S, tau
#   @unpack atmosphere = equation
#
#   ρatm = atmosphere.rho
#   patm = 0.0
#   ϵatm = tau / ρatm
#   hatm = 1.0
#   Watm = 1.0
#   vatm = 1.0
#
#   if rho < atmosphere.rho * atmosphere.threshold
#     rho = rhoatm
#     p   = patm
#     eps = epsatm
#     v   = vatm
#     D   = Formulae.D(rhoatm, vatm, epsatm, patm)
#     S   = Formulae.S(rhoatm, vatm, epsatm, patm)
#     tau = Formulae.τ(rhoatm, vatm, epsatm, patm)
#   end
#
#   @returns rho, p, eps, v, D, S, tau
# end


@with_signature function primitives(equation::Equation)
  @accepts D, S, tau, p
  rho = Formulae.ρ(D, S, tau, p)
  v   = Formulae.v(D, S, tau, p)
  eps = Formulae.ϵ(D, S, tau, p)
  @returns rho, v, eps
end


@with_signature function conservatives(equation::Equation)
  @accepts rho, v, eps, p
  D   = Formulae.D(rho, v, eps, p)
  S   = Formulae.S(rho, v, eps, p)
  tau = Formulae.τ(rho, v, eps, p)
  @returns D, S, tau
end


@with_signature function entropy_variables(equation::Equation{EquationOfState.IdealGas})
  @accepts rho, p, E, flx_E
  @unpack eos = equation
  γ     = eos.gamma
  E     = rho / (γ - 1) * log(p / rho^γ)
  flx_E = q * E / rho
  @returns E, flx_E
end


@with_signature function cons2prim(equation::Equation)
  @accepts D, S, tau, p, x
  rho, v, eps, p = cons2prim_insane((D, S, tau, p, x), equation)
  @returns rho, v, eps, p
end


@with_signature function cons2prim_insane(equation::Equation)
  @accepts D, S, tau, p, x
  (p,) = find_pressure_insane((D, S, tau, p, x), equation)
  rho  = Formulae.ρ(D, S, tau, p)
  v    = Formulae.v(D, S, tau, p)
  eps  = Formulae.ϵ(D, S, tau, p)
  @returns rho, v, eps, p
end


function pressure_equation_insane(tau, p, D, S, eos, p0, x)
  S2 = S^2
  a = tau + p + D
  a2mS2 = a^2 - S2
  if a2mS2 < 0
    return NaN
  end
  W = a / sqrt(a2mS2)
  rho = D / W
  eps = (a - W^2 * p) / (W^2 * rho) - 1
  if eps < 0 # will break pressure computation below
    return NaN
  end
  eos_p = eos(Pressure, Density, rho, InternalEnergy, eps)
  f = eos_p - p
  return f
end


@with_signature function find_pressure_insane(equation::Equation)
  @accepts D, S, tau, p, x
  @unpack eos = equation

  # upper limit on particle velocity
  omv2 = 1e-3 # = 1 - v^2

  S2   = S^2
  # fails in relativistic shocktube 2
  # if tau <= -1.0 || D <= 0.0
  #   error("Invalid state encountered! (τ, D) = ($tau, $D)")
  # end
  pmin1 = - (tau + D) + sqrt(S2 + omv2) # √((τ + D + p)^2 - S^2) >= √(1 - vmax^2)
  pmin2 = 0 # physical constraint
  pmin = max(pmin1, pmin2)

  floor_pressure(pnew) = max(pnew, pmin)
  # p0 = 0.5 * (eos.gamma - 1) * (tau + D)
  p0 = p

  f = pp -> pressure_equation_insane(tau, pp, D, S, eos, p0, x)
  @unpack cons2prim_newtonsolver, cons2prim_bisection = equation
  p, err, converged, tol = find_root(f, p0, cons2prim_newtonsolver, sanitize=floor_pressure)
  if !converged
    p = pmin
  end

  @returns p
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_flux(equations::Equation)
  @accepts State(D, S, tau)
  flx_ldg_D, flx_ldg_S, flx_ldg_tau = D, S, tau
  @returns flx_ldg_D, flx_ldg_S, flx_ldg_tau
end


@with_signature function ldg_speed(equations::Equation)
  @accepts State(D, S, tau)
  v_ldg_D, v_ldg_S, v_ldg_tau = 1, 1, 1
  @returns v_ldg_D, v_ldg_S, v_ldg_tau
end


@with_signature function av_flux(equations::Equation)
  @accepts State(ldg_D, ldg_S, ldg_tau), smoothed_mu
  flx_ldg_D, flx_ldg_S, flx_ldg_tau =
    smoothed_mu * ldg_D, smoothed_mu * ldg_S, smoothed_mu * ldg_tau
  @returns flx_ldg_D, flx_ldg_S, flx_ldg_tau
end


@with_signature function av_speed(equations::Equation)
  @accepts State(ldg_D, ldg_S, ldg_tau), smoothed_mu
  flx_ldg_D, flx_ldg_S, flx_ldg_tau = smoothed_mu, smoothed_mu, smoothed_mu
  @returns v_flx_ldg_D, v_flx_ldg_S, v_flx_ldg_tau
end
