function Project(env::Environment, prms)

  # construct Project
  equation = make_Equation(prms["ScalarEq"])
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])
  rsolver = ApproxRiemannSolver(flux, speed, equation)
  ldg_rsolver, av_rsolver = if hrsc isa AbstractArtificialViscosity
    ApproxRiemannSolver(ldg_flux, ldg_speed, equation),
    ApproxRiemannSolver(av_flux, av_flux, equation)
  else
    nothing, nothing
  end

  P = Project(equation, rsolver, hrsc, tci, ldg_rsolver, av_rsolver)

  # register variables
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(cache, equation)
  _register_variables!(cache, tci)
  _register_variables!(cache, hrsc)
  !isnothing(rsolver) && register_variables!(cache, rsolver)
  !isnothing(ldg_rsolver) && register_variables!(cache, ldg_rsolver)
  !isnothing(av_rsolver) && register_variables!(cache, av_rsolver, dont_register=true)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["ScalarEq"])

  # setup boundary conditions
  bdryconds = make_BoundaryConditions(env, equation, rsolver, prms["ScalarEq"])
  ldg_bdryconds, av_bdryconds = if hrsc isa AbstractArtificialViscosity
    make_BoundaryConditions_LDG(env, equation, ldg_rsolver, prms["ScalarEq"]),
    make_BoundaryConditions_LDG(env, equation, av_rsolver, prms["ScalarEq"])
  else
    nothing, nothing
  end

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup callbacks
  projectcb = make_callback(env, P, bdryconds)

  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


function _register_variables!(cache, tci) end


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
# Perhaps this is the import vs. using confusion people talk on the forums.


function _register_variables!(cache, eq::AbstractScalarEq)
  register_variables!(cache,
    dynamic_variablenames = (:u,),
    rhs_variablenames     = (:rhs_u,),
    static_variablenames  = (:flx_u, :E, :Em1, :F, :Fm1, :v),
    cell_variablenames    = (:max_v,),
    global_variablenames  = (:t, :tm1))
end


function _register_variables!(cache, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(cache,
    static_variablenames  = (:q, :flx_q),
    cell_variablenames    = (:mu,))
end


function _register_variables!(cache, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(cache, hrsc.av)
  register_variables!(cache, static_variablenames  = (:smoothed_mu,))
end


function _register_variables!(cache, hrsc::Maybe{HRSC.AbstractHRSC}) end


function _register_variables!(cache, tci::TCI.AbstractTCI)
  register_variables!(cache, cell_variablenames = (:flag,))
end


function _register_variables!(cache, tci::TCI.EntropyProduction)
  register_variables!(cache,
                      cell_variablenames   = (:flag,),
                      static_variablenames = (:EP,),
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs))
end
