# TODO
# - rename n_points, n_cells to npoints, ncells
# - change Mesh.x, Mesh.invjac from a Matrix to a Vector
# - rename Mesh to Mesh1d


#######################################################################
#                                Types                                #
#######################################################################


"""
    Mesh

Object that bundles grid and operators needed for modeling a Discontinuous Galerkin 
scheme in 1D.
"""
struct Mesh
  element::SpectralElement  # collection of spectral element operators
  K::Int64                  # number of cells
  xmin::Float64             # left most grid coordinate
  xmax::Float64             # right most grid coordiante
  CFL::Float64              # Courant Friedrichs Lewy factor
  L::Float64                # domain length
  dl::Float64               # equidistant subcell length
  x::Matrix{Float64}        # grid points
  invjac::Matrix{Float64}   # inverse jacobian of coordinate transformation
                            # from computational [-1,1] onto physical grid [xmin,xmax]
end


function Mesh(; N=5, K=10, xrange=[-1.0,1.0], CFL=0.5, basis="lgl_lumped")

  @toggled_assert N > 0
  @toggled_assert K > 0
  @toggled_assert CFL > 0
  @toggled_assert length(xrange) == 2
  @toggled_assert xrange[1] < xrange[2]
  @toggled_assert basis in ["lgl", "glgl", "lgl_lumped"]

  # TODO Remove this after updating parameter names in SpectralElement
  quadrature_method = Dict("lgl" => :LGL, "glgl" => :GLGL, "lgl_lumped" => :LGL_lumped)[basis]

  element = SpectralElement(N, Symbol(quadrature_method))
  @unpack z = element
  xmin, xmax = xrange
  L = xmax - xmin
  dl = L/K
  a = [ xmin + dl * idx for idx = 0:K-1 ]
  b = a .+ dl
  Npts = length(z)
  x = Matrix{Float64}(undef, Npts, K)
  invjac = similar(x)
  for i=1:Npts, j=1:K
    aj = a[j]
    bj = b[j]
    zi = z[i]
    x[i,j] = 0.5 * ((bj - aj) * zi + (aj + bj))
    invjac[i,j] = 2.0 / (bj - aj)
  end
  # sanitize cell end points to have same x coordinate at common interfaces
  for j=1:K-1
    xl = x[end,j]
    xr = x[1,j+1]
    xmean = 0.5 * (xl + xr)
    x[end,j] = xmean
    x[1,j+1] = xmean
  end

  return Mesh(element, K, xmin, xmax, CFL, L, dl, x, invjac)
end


#######################################################################
#                               Methods                               #
#######################################################################


Base.broadcastable(mesh::Mesh) = Ref(mesh)


"""
  extract_cell_bdrys(mesh::Mesh)

Return coordinates of cell boundaries from a `mesh`.
"""
function extract_cell_bdrys(mesh::Mesh)
  @unpack x, K = mesh
  @unpack Npts = mesh.element
  xbs = Float64[]
  for k = 1:K
    push!(xbs, x[(k-1)*Npts+1])
  end
  push!(xbs, x[end])
  return xbs
end


"""
  grid_average(u, mesh::Mesh)

Return the grid average of all cell averages 
"""
function grid_average(u, mesh::Mesh)
  @unpack element, dl, L = mesh
  cellavg = cellwise_average(u, mesh) .* dl
  avg = sum(cellavg) / L
  return avg
end


"""
  cellwise_average(u, mesh::Mesh)

Return the cellwise average of `u` over `mesh.x`.
"""
function cellwise_average(u, mesh::Mesh)
  @unpack element, invjac, K, dl = mesh
  @unpack Npts = element
  ru = vreshape(u, (Npts, K))
  avg = zeros(K)
  @views for i in 1:K
    avg[i] = integrate(ru[:,i], element)
  end
  avg ./= invjac[1,:] * dl # constant Jacobian, constant cell width
  return avg
end


function inner_product(u, v, k, mesh::Mesh)
  @unpack element, invjac, K = mesh
  @unpack Npts = element
  ru = vreshape(u, (Npts,K))
  rv = vreshape(v, (Npts,K))
  @views iprod = inner_product(ru[:,k], rv[:,k], element)
  iprod /= invjac[1,k]
  return iprod
end


"""
  cellwise_inner_product(u, v, mesh::Mesh)
  cellwise_inner_product!(uv, u, v, mesh::Mesh)

Return vector of inner products between `u, v` on every cell in `mesh`.
"""
function cellwise_inner_product(u, v, mesh::Mesh)
  uv = zeros(Float64, size(u))
  cellwise_inner_product!(uv, u, v, mesh)
  return uv
end
function cellwise_inner_product!(uv, u, v, mesh::Mesh)
  @unpack element, K, invjac = mesh
  @unpack Npts = element
  L = layout(mesh)
  ru = vreshape(u, L)
  rv = vreshape(v, L)
  for k = 1:K
    uk = view(ru, :, k)
    vk = view(rv, :, k)
    uv[k] = inner_product(uk, vk, element) / invjac[1,k]
  end
  return
end


"""
  broken_inner_product(u, v, mesh::Mesh)

Return sum of all inner products between `u,v` on every cell in `mesh`.
"""
function broken_inner_product(u, v, mesh::Mesh)
  inner_products = cellwise_inner_product(u, v, mesh)
  return sum(inner_products)
end


function LinearAlgebra.norm(u, mesh::Mesh, order)
  if order == 1
    return norm_L1(u, mesh)
  elseif order == 2
    return norm_L2(u, mesh)
  else
    error("Norm for order = $order not implemented.")
  end
end
norm_L2(u, mesh::Mesh) = sqrt(broken_inner_product(u, u, mesh))
# TODO Refine broken_inner_product and other mesh quadrature methods to also accept
# a function that can be applied somewhere in the underlying integrate function before
# the integrand is evaluated. This should help avoid making a copy of the array here.
norm_L1(u, mesh::Mesh) = broken_inner_product(abs.(u), ones(Float64, size(u)), mesh)


"""
    find_nearest(xi, mesh::Mesh)

Find index of node in `mesh` that lies closest to `xi`.
Returns the index of the outermost point if `xi` falls outside `mesh`.
"""
function find_nearest(xi, mesh::Mesh)
  @unpack x = mesh
  _, idx = findmin(xx -> abs(xx - xi), view(x,:))
  if isnothing(idx)
    return xi < mesh.xmin ? 1 : length(mesh)
  end
  return idx
end


@deprecate locate_point(xi, mesh::Mesh) find_nearest(xi, mesh::Mesh)


"""
    find_cell_index(xi, mesh::Mesh)

Find the cell index into which coordinate `xi` falls into the `mesh`.
Note that it returns the first index found in case `xi` falls onto an interface boundary.
Returns nothing if `xi` lies outside the `mesh`.
"""
function find_cell_index(xi, mesh::Mesh)
  cell_bdrys = extract_cell_bdrys(mesh)
  cell_index = findfirst(idx -> cell_bdrys[idx] <= xi <= cell_bdrys[idx+1], 1:length(cell_bdrys)-1)
  return cell_index
end


Base.length(mesh::Mesh)::Int64 = length(mesh.x)
Base.extrema(mesh::Mesh) = extrema(mesh.x)
n_points(mesh::Mesh)::Int = mesh.element.Npts
n_cells(mesh::Mesh) = mesh.K
layout(mesh::Mesh)::Tuple{Int,Int} = (n_points(mesh), n_cells(mesh))


"""
    MeshInterpolator


Operator for interpolating data from one mesh to another.

# Examples

```julia
julia> sample_mesh, reference_mesh = Mesh(; N=3, K=4); Mesh(; N=10, K=20);

julia> reference_data = sin.(reference_mesh.x);

julia> meshintrp = MeshInterpolator(sample_mesh, reference_mesh);

julia> interp_data = meshintrp(reference_data)

julia> meshintrp(intrp_data, reference_data) # result is stored inside intrp_data
```

Note that for inplace interpolation `intrp_data` and `reference_data` cannot be aliased.
"""
struct MeshInterpolator
  reference_mesh::Mesh
  sample_mesh::Mesh
  interpolation_matrices::Vector{Tuple{Int64,Matrix{Float64}}}
  reference_cell_indices::Vector{Int64}
end


function MeshInterpolator(smpl_mesh::Mesh, ref_mesh::Mesh)

  smpl_xmin, smpl_xmax = extrema(smpl_mesh)
  ref_xmin, ref_xmax   = extrema(ref_mesh)
  if !(ref_xmin <= smpl_xmin <= ref_xmax) || !(ref_xmin <= smpl_xmax <= ref_xmax)
    error("Sample mesh must fall into reference mesh")
  end

  smpl_x = vreshape(smpl_mesh.x, layout(smpl_mesh))
  ref_x  = vreshape(ref_mesh.x, layout(ref_mesh))

  ref_xbdrys       = extract_cell_bdrys(ref_mesh)
  ref_cell_indices = find_cell_index.(smpl_x, Ref(ref_mesh))

  interpolation_matrices = Tuple{Int64, Matrix{Float64}}[]
  current_idx = 1 # first index of sample node that falls into a new cell in the reference mesh
  while true
    cell_idx  = ref_cell_indices[current_idx]
    # find all sample indices of nodes which fall into the same reference cell
    same_idxs = findall(ref_idx -> ref_idx == cell_idx, ref_cell_indices)
    smpl_idxs = [ current_idx ]
    if length(same_idxs) == 1
      append!(smpl_idxs, same_idxs[2:end])
    end
    # extract the sample nodes
    smpl_xs   = smpl_x[smpl_idxs]
    # generate interpolation matrix
    push!(interpolation_matrices,
          (cell_idx, dg1d.lagrange_interp_matrix(smpl_xs, ref_x[:,cell_idx])))

    current_idx += length(smpl_idxs)
    if current_idx > length(ref_cell_indices)
      break
    end
  end

  return MeshInterpolator(ref_mesh, smpl_mesh, interpolation_matrices, ref_cell_indices[:])
end


function (intrp::MeshInterpolator)(ref_data)
  result = zeros(layout(intrp.sample_mesh))
  intrp(result, ref_data)
  return result
end


function (intrp::MeshInterpolator)(storage, ref_data)

  if length(ref_data) != length(intrp.reference_mesh)
    error("Provided reference data does not match with reference mesh.")
  end
  if length(storage) != length(intrp.sample_mesh)
    error("Provided storage for interpolated result does not match with sample mesh.")
  end

  vec_intrp_data = view(storage, :)
  mat_ref_data   = vreshape(ref_data, layout(intrp.reference_mesh))
  current_idx = 1 # first index of sample node that falls into a new cell in the reference mesh
  @views for (ref_cell_idx, intrp_mat) in intrp.interpolation_matrices
    n_intrp_data = size(intrp_mat)[1]
    idx_range = current_idx:current_idx+(n_intrp_data-1)
    vec_intrp_data[idx_range] .= intrp_mat * mat_ref_data[:, ref_cell_idx]
    current_idx += n_intrp_data
  end

  return
end
