"""
    abstract AbstractRiemannSolver 

Supertype for `RiemannSolver` interface.
Any Subtype `Solver <: RiemannSolver` must implement the following methods
- `register_variables!(cache, rsolver::Solver)`,
- `broadcast_faces!(f, rsolver::Solver, cache::Cache, mesh::Mesh)` where `f` is a 
  `@with_signature` function.
- `broadcast_boundaries!(f, rsolver::Solver, cache::Cache, mesh::Mesh)` where `f` is a 
  `@with_signature` function.

See: `ApproxRiemannSolver`
"""
abstract type AbstractRiemannSolver end


"""
    ApproxRiemannSolver <: AbstractRiemannSolver

Approximate Riemann solver interface for numerical flux computation.
Available implementations: `lax_friedrich_flux`, `central_flux`.

---

    ApproxRiemannSolver(flux, speed, equation; lhs_bc=PeriodicBC(), rhs_bc=PeriodicBC())

Assumes an `equation::SomeEquation` that implements two `@with_signature` functions
- `flux` - compute fluxes given a state,
- `speed` - compute maximum wave speed given a state.

# Example

```julia
julia> mesh = Mesh(); cache = Cache(mesh);

julia> rsolver = ApproxRiemannSolver(flux, speed, equation)

# register numerical fluxes as bdry variables in cache
julia> register_variables!(cache, rsolver);

# compute numerical fluxes inplace in cache
julia> broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
```

"""
struct ApproxRiemannSolver{T_Flux<:Function,
                           T_Speed<:Function,
                           T_Equation<:AbstractEquation,
                           N,M} <: AbstractRiemannSolver

  flux::T_Flux
  speed::T_Speed
  equation::T_Equation
  accepts::NTuple{N,Symbol}
  returns::Tuple{NTuple{M,Symbol},NTuple{M,Symbol}}

  function ApproxRiemannSolver(flux, speed, equation)
    @assert has_signature(flux, equation) begin
      "ApproxRiemannSolver: Require a @with_signature function for '$flux'"
    end
    @assert has_signature(speed, equation) begin
      "ApproxRiemannSolver: Require a @with_signature function for '$speed'"
    end
    accepts_flux, returns_flux  = signature(flux, equation)
    accepts_speed = accepts(speed, equation)
    state_indices_flux = state_indices(flux, equation)
    state_indices_speed = state_indices(speed, equation)
    @assert all(accepts_flux .=== accepts_speed) && state_indices_flux == state_indices_speed begin
      "ApproxRiemannSolver: '$flux' and '$speed' must accept the same variables"
    end
    rsolve_accepts = accepts_flux
    rsolve_returns = (Symbol.(:lhs_num, returns_flux), Symbol.(:rhs_num, returns_flux))
    return new{typeof(flux), typeof(speed), typeof(equation),
               length(accepts_flux), length(returns_flux)}(flux, speed, equation,
                                                           rsolve_accepts, rsolve_returns)
  end
end


# handmade definitions for @with_signature flux functions
# - accepts the same variables as rsolve.flux, rsolve.speed
# - state_indices are taken from rsolve.flux, rsolve.speed
# - returns a tuple of tuples of numerical fluxes, e.g. ( (lhs_numflx_u,), (rhs_numflx_u,) )
#   this is different from the standard @with_signature interface, but we need it to separate
#   the StructArrays belonging to the numerical fluxes on the left and right interfaces.
state_indices(f, rsolve::ApproxRiemannSolver) = state_indices(rsolve.flux, rsolve.equation)
function signature(f, rsolve::ApproxRiemannSolver)
  return rsolve.accepts, rsolve.returns
end


state(tpl, rsolve::ApproxRiemannSolver) = state(tpl, state_indices(rsolve.flux, rsolve.equation))


function register_variables!(cache, rsolver::ApproxRiemannSolver; dont_register=false)
  @unpack flux, equation = rsolver
  returns_vars_lhs, returns_vars_rhs = returns(lax_friedrich_flux, rsolver)
  if !dont_register
    register_variables!(cache, bdry_variablenames=returns_vars_lhs)
    register_variables!(cache, bdry_variablenames=returns_vars_rhs)
  end
end


#######################################################################
#             Approximate Riemann Solver implementations              #
#######################################################################


"""
    @with_signature lax_friedrich_flux(args, rsolver::ApproxRiemannSolver)

See: `ApproxRiemannSolver`
"""
function lax_friedrich_flux(args, rsolver::ApproxRiemannSolver)
  args_lhs, args_rhs = args
  @unpack flux, speed, equation = rsolver
  ul, ur = state(args_lhs, rsolver), state(args_rhs, rsolver)
  vl, vr = speed(args_lhs, equation), speed(args_rhs, equation)
  fl, fr = flux(args_lhs, equation), flux(args_rhs, equation)
  C = max(abs.(vl), abs.(vr))
  nf = @. 0.5 * (fl + fr) + 0.5 * C * (ul - ur)
  (nf, nf)
end
has_signature(::typeof(lax_friedrich_flux), ::ApproxRiemannSolver) = true


"""
    @with_signature lax_friedrich_flux(args, rsolver::ApproxRiemannSolver)

See: `ApproxRiemannSolver`
"""
function avg_lax_friedrich_flux(args, rsolver::ApproxRiemannSolver)
  args_lhs, args_rhs = args
  @unpack flux, speed, equation = rsolver
  ul, ur = state(args_lhs, rsolver), state(args_rhs, rsolver)
  vl, vr = speed(args_lhs, equation), speed(args_rhs, equation)
  avg_args = @. 0.5 * (args_lhs + args_rhs)
  avg_f = flux(avg_args, equation)
  # fl, fr = flux(args_lhs, equation), flux(args_rhs, equation)
  C = max(abs.(vl), abs.(vr))
  nf = @. avg_f + 0.5 * C * (ul - ur)
  (nf, nf)
end
has_signature(::typeof(avg_lax_friedrich_flux), ::ApproxRiemannSolver) = true


"""
    @with_signature central_flux(args, rsolver::ApproxRiemannSolver)

See: `ApproxRiemannSolver`
"""
function central_flux(args, rsolver::ApproxRiemannSolver)
  args_lhs, args_rhs = args
  ul, ur = state(args_lhs, rsolver), state(args_rhs, rsolver)
  nf = @. 0.5 * (ul + ur)
  (nf, nf)
end
has_signature(::typeof(central_flux), ::ApproxRiemannSolver) = true


"""
    @with_signature mean_flux(args, rsolver::ApproxRiemannSolver)

See: `ApproxRiemannSolver`
"""
function mean_flux(args, rsolver::ApproxRiemannSolver)
  args_lhs, args_rhs = args
  @unpack flux, equation = rsolver
  fl, fr = flux(args_lhs, equation), flux(args_rhs, equation)
  nf = @. 0.5 * (fl + fr)
  (nf, nf)
end
has_signature(::typeof(mean_flux), ::ApproxRiemannSolver) = true
