"""
    @with_signature

Macro to extract `accepted` and `returned` arguments from a function and make them
programmatically accessible. Functions defined using this macro can be used with
the `broadcast_volume!, broadcast_faces!, broadcast_bdrys` interfaces.


# Example
```
@with_signature function fn(eq::Equation)
  @accepts State(u1, u2), a, b
  # compute values rhs_u1, rhs_u2
  @returns rhs_u1, rhs_u2
end
```
translates into
```
function fn(args::NTuple{4,Float64}, eq::Equation)
  (u1, u2, a, b) = args
  # compute values rhs_u1, rhs_u2
  (rhs_u1, rhs_u2)
end
accepts(::typeof(fn), ::Equation) = (:u1, :u2, :a, :b)
returns(::typeof(fn), ::Equation) = (:rhs_u1, :rhs_u2)
state_indices(::typeof(fn), ::Equation) = (1,2)
has_signature(::typeof(fn), ::Equation) = true
```
"""
macro with_signature(ex)
  return with_signature(ex)
end


"""
See @with_signature
"""
macro accepts(ex); ex; end
macro returns(ex); ex; end


_is_state(ex) = ex isa Expr && ex.head === :call && ex.args[1] === :State
_is_tuple(ex) = ex isa Expr && ex.head === :tuple


# gather list of all variables and a list of indices of variables
# tagged with State, e.g.
# accepts(:(State(a, b), c, d, State(e), f))
# ->
# (:a, :b, :c, :d, :e, :f), (1, 2, 5)
function accepts(ex)
  all, i_states = if ex isa Symbol
    (ex,), ()
  elseif _is_state(ex)
    vars = ex.args[2:end]
    vars, collect(1:length(vars))
  elseif _is_tuple(ex)
    vars = ex.args[1:end]
    all, i_states = Symbol[], Int64[]
    last_i = 1
    for v in vars
      if _is_state(v)
        vs = v.args[2:end]
        push!(all, vs...)
        append!(i_states, collect(1:length(vs)) .+ (last_i-1))
        last_i += length(vars)-1
      elseif v isa Symbol
        push!(all, v)
      else
        error("@accepts: expected a variable name or a State tag")
      end
    end
    all, i_states
  else
    error("@accepts: expected '@accepts a, b, ...'")
  end
  tuple(all...), tuple(i_states...)
end


function returns(ex)
  if ex isa Symbol
    (ex,)
  elseif _is_tuple(ex)
    tuple(ex.args...)
  else
    error("@returns: expected '@returns f_a, f_b, ...'")
  end
end


state(tpl, is::NTuple{N,Int64}) where N = getindex.(Ref(tpl), is)


function with_signature(ex)
  ex = MacroTools.prewalk(MacroTools.rmlines, ex)
  # display(ex)

  ex.head === :function || error("@with_signature: only works with functions")
  fn_signature = ex.args[1]
  length(fn_signature.args) == 2 || error(
    "@with_signature: function with exactly one argument required")
  fn_name, first_arg = fn_signature.args

  ex.args[2].head === :block || error("@with_singature: failed to extract function body")
  fn_body = ex.args[2].args

  # extract place holders
  length(fn_body) == 0 && error("@with_signature: missing function body")
  matched_accepts = MacroTools.@capture(fn_body[1], @accepts accepts_variables_)
  matched_accepts || error(
    "@with_signature: expected '@accepts a, b, ...' at the start of the function body")
  for line in fn_body[2:end]
    MacroTools.@capture(line, @accepts rest_) && error(
    "@with_signature: @accepts can only occur once")
  end
  matched_returns = MacroTools.@capture(fn_body[end], @returns returns_variables_)
  matched_accepts || error(
    "@with_signature: expected '@returns a, b, ...' at the end of the function body")
  for line in fn_body[1:end-1]
    MacroTools.@capture(line, @returns rest_) && error(
      "@with_signature: @returns can only occur once")
  end

  # gather variables from place holders
  accepts_variables, pos_state_variables = accepts(accepts_variables)
  returns_variables = returns(returns_variables)

  fn_body[1]   = Expr(:(=), Expr(:tuple, accepts_variables...), :args)
  fn_body[end] = Expr(:tuple, returns_variables...)

  # sanitize function arguments
  (first_arg isa Expr && first_arg.head === :(::)) || error(
    "@with_signature: function argument '$first_arg' must be type annotated")
  dispatch_type = last(first_arg.args)

  # values are passed via an additional argument
  # we type annotate it here with a ::NTuple{N,Float64}
  n_accept_variables = length(accepts_variables)
  annotated_variables_arg = :(args::NTuple{$n_accept_variables,Float64})
  new_fn_signature = Expr(:call, fn_name, annotated_variables_arg, first_arg)
  ex.args[1] = new_fn_signature

  thismod = @__MODULE__

  new_ex = esc(quote
    $ex
    $(thismod).signature(::typeof($(fn_name)), ::$(dispatch_type)) = 
        ($(accepts_variables), $(returns_variables))
    $(thismod).state_indices(::typeof($(fn_name)), ::$(dispatch_type)) = $(pos_state_variables)
    $(thismod).has_signature(::typeof($(fn_name)), ::$(dispatch_type)) = true
  end)

  # display(MacroTools.prewalk(MacroTools.rmlines, new_ex))

  return new_ex
end


#######################################################################
#                          Interface methods                          #
#######################################################################


has_signature(fn, dispatch) = false
signature(fn, dispatch) = error(
  "function '$fn' with dispatch '$(dispatch)' was not defined using @with_signature")
state_indices(fn, dispatch) = error(
  "function '$fn' with dispatch '$(dispatch)' was not defined using @with_signature")
accepts(fn, dispatch) = signature(fn, dispatch)[1]
returns(fn, dispatch) = signature(fn, dispatch)[2]


#######################################################################
#                          broadcast methods                          #
#######################################################################


"""
    broadcast_volume!(f::Function, D::T_Dispatch, C::Cache) where T_Dispatch

Broadcast a `@with_signature` function `f` dispatched on `D` over variables of
cache `C`. The arguments and return values are determined by the `accepts, returns` overloads
of `(f, D)` from the `@with_signature` interface.

See: `@with_signature`.
"""
function broadcast_volume!(f::Function, D::T_Dispatch, C::Cache) where T_Dispatch
  has_signature(f, D) || error("broadcast_volume!: Function '$f' has no signature!")
  accepts_syms, returns_syms = accepts(f, D), returns(f, D)
  n_accepts, n_returns = length(accepts_syms), length(returns_syms)
  v = get_variable(C, first(accepts_syms))
  accepts_vars = NTuple{n_accepts,Vector{Float64}}( get_variable(C, a) for a in accepts_syms )
  returns_vars = NTuple{n_returns,Vector{Float64}}( get_variable(C, r) for r in returns_syms )
  f_args = args -> f(args, D)
  _broadcast_volume!(f_args, C, accepts_vars, returns_vars)
  return
end


function _broadcast_volume!(f, C::Cache,
    accepts_vars::NTuple{N,Vector{Float64}},
    returns_vars::NTuple{M,Vector{Float64}}) where {N,M}

  soa_accepts = StructArray{NTuple{N,Float64}}( accepts_vars )
  soa_returns = StructArray{NTuple{M,Float64}}( returns_vars )
  npts = length(first(accepts_vars))
  for idx = 1:npts
    soa_returns[idx] = f(soa_accepts[idx])
  end
  return
end


@deprecate broadcast_volume!(f, rsolver, cache, mesh) broadcast_faces!(f, rsolver, cache, mesh)


"""
    broadcast_faces!(f, rsolver::ApproxRiemannSolver, cache::Cache, mesh::Mesh)

Broadcast `@with_signature` function `f` with dispatch `rsolver` over `cache` and
faces of `mesh`.
"""
function broadcast_faces!(f, rsolver::ApproxRiemannSolver, cache::Cache, mesh::Mesh)
  has_signature(f, rsolver) || error("broadcast_volume!: Function '$f' has no signature!")
  accepts_vars = accepts(f, rsolver)
  returns_vars_lhs, returns_vars_rhs = returns(f, rsolver)
  n_accepts     = length(accepts_vars)
  n_returns_lhs = length(returns_vars_lhs)
  n_returns_rhs = length(returns_vars_rhs)
  accepts_data     = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                       for a in accepts_vars )
  returns_lhs_data = NTuple{n_returns_lhs,Vector{Float64}}( get_variable(cache, r)
                                                           for r in returns_vars_lhs )
  returns_rhs_data = NTuple{n_returns_rhs,Vector{Float64}}( get_variable(cache, r)
                                                           for r in returns_vars_rhs )
  f_args = args -> f(args, rsolver)
  _broadcast_face!(f_args, cache, mesh, accepts_data, returns_lhs_data, returns_rhs_data)
  return
end


# similar to _broadcast_volume!, but only iterating cell bdrys and broadcasting
# over two returned data arrays
function _broadcast_face!(f, cache::Cache, mesh::Mesh,
    accepts_data::NTuple{Na,Vector{Float64}},
    returns_lhs_data::NTuple{Nr,Vector{Float64}},
    returns_rhs_data::NTuple{Nr,Vector{Float64}}) where {Na,Nr}

  soa_accepts = StructArray{NTuple{Na,Float64}}( accepts_data )
  soa_returns_lhs = StructArray{NTuple{Nr,Float64}}( returns_lhs_data )
  soa_returns_rhs = StructArray{NTuple{Nr,Float64}}( returns_rhs_data )
  Npts, K = layout(mesh)
  Idx = LinearIndices((Npts, K))
  for cell = 1:K
    cell_lhs, cell_rhs = periodic_index(cell-1, K), periodic_index(cell, K)
    idx_lhs, idx_rhs = Idx[end,cell_lhs], Idx[1,cell_rhs]
    accepts_lhs, accepts_rhs = soa_accepts[idx_lhs], soa_accepts[idx_rhs]
    returns_lhs, returns_rhs = f((accepts_lhs, accepts_rhs))
    soa_returns_lhs[cell_rhs] = returns_lhs
    soa_returns_rhs[cell_lhs] = returns_rhs
  end
  return
end



"""
    broadcast_boundaryconditions!(f, bc::BoundaryConditions, cache::Cache, mesh::Mesh, t)

Impose boundary conditions using the flux method `f` and boundary conditions `bc`
for fields in `cache` on `mesh` at time `t`.
"""
function broadcast_boundaryconditions!(f, bc::BoundaryConditions, cache::Cache, mesh::Mesh, t)
  @unpack lhs_bc, rhs_bc, rsolver = bc
  has_signature(f, rsolver) || error("broadcast_boundary!: Function '$f' has no signature!")
  bc.isperiodic && return

  accepts_vars = accepts(f, bc.rsolver)
  returns_vars_lhs, returns_vars_rhs = returns(f, bc.rsolver)
  n_accepts     = length(accepts_vars)
  n_returns_lhs = length(returns_vars_lhs)
  n_returns_rhs = length(returns_vars_rhs)
  accepts_data     = NTuple{n_accepts,Vector{Float64}}( get_variable(cache, a)
                                                       for a in accepts_vars )
  returns_lhs_data = NTuple{n_returns_lhs,Vector{Float64}}( get_variable(cache, r)
                                                           for r in returns_vars_lhs )
  returns_rhs_data = NTuple{n_returns_rhs,Vector{Float64}}( get_variable(cache, r)
                                                           for r in returns_vars_rhs )
  f_args = args -> f(args, rsolver)

  _broadcast_boundaries!(f_args, cache, mesh, accepts_data, returns_lhs_data, returns_rhs_data,
                         lhs_bc, rhs_bc, t)

  return
end


boundary_state(bc::AbstractBC, u_inner, t) = bc(t)
boundary_state(bc::OutflowBC, u_inner, t)  = u_inner


# similar to _broadcast_volume!, but only iterating domain bdrys and broadcasting
# over two returned data arrays
function _broadcast_boundaries!(f, cache::Cache, mesh::Mesh,
    accepts_vars::NTuple{Na,Vector{Float64}},
    returns_lhs_data::NTuple{Nr,Vector{Float64}},
    returns_rhs_data::NTuple{Nr,Vector{Float64}},
    lhs_bc::AbstractBC,
    rhs_bc::AbstractBC,
    t) where {Na,Nr}

  soa_accepts = StructArray{NTuple{Na,Float64}}( accepts_vars )
  soa_returns_lhs = StructArray{NTuple{Nr,Float64}}( returns_lhs_data )
  soa_returns_rhs = StructArray{NTuple{Nr,Float64}}( returns_rhs_data )
  Npts, K = layout(mesh)
  Idx = LinearIndices((Npts, K))

  # lhs bdry
  cell = Idx[1,1]
  state = soa_accepts[cell]
  bdry_state = boundary_state(lhs_bc, state, t)
  _, returns_rhs = f((bdry_state, state))
  soa_returns_lhs[1] = returns_rhs

  # rhs bdry
  cell = Idx[end,end]
  state = soa_accepts[cell]
  bdry_state = boundary_state(rhs_bc, state, t)
  returns_lhs, _ = f((state, bdry_state))
  soa_returns_rhs[end] = returns_lhs

  return
end
